/* #include "graphics.h" */
/* #include <graph3d.h>  */
#include <math.h>
#include <stdio.h>

/*--- Projection Definition section -------------------------*/

static double observer[4],target[4];
static double eyedist=1;
static double normal[4],e1[4],e2[4];
static double skewness=0;
static double swidth=1.0,sheight=1.0;
static double viewx=80*M_PI/180.0,viewy=60*M_PI/180.0;
static double A[4][4],AL[4][4],AR[4][4];
static double SCALEX=1,SCALEY=1;

static int dpymode=MONOSCOPICDPY;
static int dmode=DWIREFRAME | DNOSHADING | DNORENDERING;

static int clipping=0;
static double clipmin=1.0,clipmax=100.0;
static int flx,fly,frx,fry;
static int LEFTCOLOR,RIGHTCOLOR,BOTHCOLOR;

void vector(v,h,x,y,z)
double v[4];
double h,x,y,z;
{
  v[0]=h;
  v[1]=x;
  v[2]=y;
  v[3]=z;
}

void SetObserver(x,y,z,skew,ed)
double x,y,z,skew,ed;
{
  observer=vector(4,1,x,y,z);
  if (skew>=-180)
    skewness=skew;
  if (ed>0)
    eyedist=ed;
  DefineProjection();
}

void SetTarget(double x,double y,double z) {
  target=vector(4,1,x,y,z);
  DefineProjection();
}

void SetSkew(double alpha) {
  skewness=alpha;
  DefineProjection();
}

void SetEyeDist(double ed) {
  eyedist=ed;
  DefineProjection();
}

void SetFieldofView(double wx,double wy) {
  viewx=wx*M_PI/180.0;
  viewy=wy*M_PI/180.0;
  DefineProjection();
}

void SetScreensize(double w,double h) {
  swidth =w;
  sheight=h;
  DefineProjection();
}

void SetFresnelPoints(int xl,int yl,int xr,int yr) {
  flx=xl;
  fly=yl;
  frx=xr;
  fry=yr;
}

void SetAnalglyphColors(int lc,int rc,int bc) {
  LEFTCOLOR =lc;
  RIGHTCOLOR=rc;
  BOTHCOLOR =bc;
}

void DefineProjection() {
  double a[4],u[4];
  vector n;
  double ex,ey,ez,el,x;

//  init_protocol("/home/schmei/pvm3/bin/graph3d.protocol");
  protocol(0);

  n=observer-target;
  normal=n/norm(n);
  u[0]=-scalar(normal,target);  // * -1 ???
  u[1]=normal.get(1);
  u[2]=normal.get(2);
  u[3]=normal.get(3);

  ex=-u[2];
  ey= u[1];
  el=ex*ex+ey*ey;
  if (el<1e-20) {
    ex=1;
    ey=0;
    ez=0;
  } else {
    el=sqrt(el);
    ex=ex/el;
    ey=ey/el;
    ez=0;
  }
  e1=rot(vector(4,0,ex,ey,ez),observer,normal,skewness);
  e2=normal*e1;
  protocol("     norm(e1)=%lf\n",norm(e1));
  protocol("     norm(e2)=%lf\n",norm(e2));
  protocol("scalar(e1,e2)=%lf\n",scalar(e1,e2));
  //--- matrix for monoscopic transformation -------------

  a[0]=1;
  a[1]=observer.get(1);
  a[2]=observer.get(2);
  a[3]=observer.get(3);
                                              /*    T  */
  x=a[0]*u[0]+a[1]*u[1]+a[2]*u[2]+a[3]*u[3];  /* = u a */
  
  A[0][0]=a[0]*u[0]-x;
  A[0][1]=a[0]*u[1];
  A[0][2]=a[0]*u[2];
  A[0][3]=a[0]*u[3];
  
  A[1][0]=a[1]*u[0];
  A[1][1]=a[1]*u[1]-x;
  A[1][2]=a[1]*u[2];
  A[1][3]=a[1]*u[3];
  
  A[2][0]=a[2]*u[0];
  A[2][1]=a[2]*u[1];
  A[2][2]=a[2]*u[2]-x;
  A[2][3]=a[2]*u[3];
  
  A[3][0]=a[3]*u[0];
  A[3][1]=a[3]*u[1];
  A[3][2]=a[3]*u[2];
  A[3][3]=a[3]*u[3]-x;

  protocol("observer=(%6.2lf,",a[1]); protocol("%6.2lf,",a[2]); protocol("%6.2lf)\n",a[3]);
  protocol("  target=(%6.2lf,",target.get(1));
  protocol("%6.2lf,",target.get(2));
  protocol("%6.2lf)\n",target.get(3));
  protocol("  normal=(%6.2lf,",u[1]); protocol("%6.2lf,",u[2]); protocol("%6.2lf)\n",u[3]);

  //--- matrix for left eye ------------------------------

  a[0]=1;
  a[1]=(observer-e1*(eyedist/2.0)).get(1);
  a[2]=(observer-e1*(eyedist/2.0)).get(2);
  a[3]=(observer-e1*(eyedist/2.0)).get(3);
  
  x=a[0]*u[0]+a[1]*u[1]+a[2]*u[2]+a[3]*u[3];
  
  AL[0][0]=a[0]*u[0]-x;
  AL[0][1]=a[0]*u[1];
  AL[0][2]=a[0]*u[2];
  AL[0][3]=a[0]*u[3];
  
  AL[1][0]=a[1]*u[0];
  AL[1][1]=a[1]*u[1]-x;
  AL[1][2]=a[1]*u[2];
  AL[1][3]=a[1]*u[3];
  
  AL[2][0]=a[2]*u[0];
  AL[2][1]=a[2]*u[1];
  AL[2][2]=a[2]*u[2]-x;
  AL[2][3]=a[2]*u[3];
  
  AL[3][0]=a[3]*u[0];
  AL[3][1]=a[3]*u[1];
  AL[3][2]=a[3]*u[2];
  AL[3][3]=a[3]*u[3]-x;

  //--- matrix for right eye -----------------------------

  a[0]=1;
  a[1]=(observer+e1*(eyedist/2.0)).get(1);
  a[2]=(observer+e1*(eyedist/2.0)).get(2);
  a[3]=(observer+e1*(eyedist/2.0)).get(3);
  
  x=a[0]*u[0]+a[1]*u[1]+a[2]*u[2]+a[3]*u[3];
  
  AR[0][0]=a[0]*u[0]-x;
  AR[0][1]=a[0]*u[1];
  AR[0][2]=a[0]*u[2];
  AR[0][3]=a[0]*u[3];
  
  AR[1][0]=a[1]*u[0];
  AR[1][1]=a[1]*u[1]-x;
  AR[1][2]=a[1]*u[2];
  AR[1][3]=a[1]*u[3];
  
  AR[2][0]=a[2]*u[0];
  AR[2][1]=a[2]*u[1];
  AR[2][2]=a[2]*u[2]-x;
  AR[2][3]=a[2]*u[3];
  
  AR[3][0]=a[3]*u[0];
  AR[3][1]=a[3]*u[1];
  AR[3][2]=a[3]*u[2];
  AR[3][3]=a[3]*u[3]-x;

  SCALEX=swidth /(2.0*norm(observer-target)*tan(viewx/2.0));
  SCALEY=sheight/(2.0*norm(observer-target)*tan(viewy/2.0));
  if (SCALEX<=SCALEY)
    SCALEY=SCALEX;
  else
    SCALEX=SCALEY;

  if (dpymode & FRESNELDPY) {
    SCALEX/=2;
    SCALEY/=2;
  }

  protocol("      e1=(%6.2lf,",e1.get(1));
  protocol("%6.2lf,",e1.get(2));
  protocol("%6.2lf)\n",e1.get(3));
  
  protocol("      e2=(%6.2lf,",e2.get(1));
  protocol("%6.2lf,",e2.get(2));
  protocol("%6.2lf)\n",e2.get(3));
  
  protocol("  SCALEX=%lf\n",SCALEX);
  protocol("  SCALEY=%lf\n",SCALEY);
//  external_cleardevice();
}


//--- Displaymode section -----------------------------------

void SetDisplayMode(int dpmode) {
  dpymode=dpmode;
  DefineProjection();
}

int GetDisplayMode() {
  return dpymode;
}

void SetClipping(int onoff,double minc,double maxc) {
  clipping=onoff;
  if ((minc>0) && (maxc>0)) {
    clipmin=minc;
    clipmax=maxc;
  }
}

//--- Transformation section --------------------------------

//--- stereoscopic transformation

int T3D2D(const vector &x,int *xl,int *yl,int *xr,int *yr) {
  vector xp;
  double d;

  if (clipping) {
    xp=x-observer;
    d=norm(xp);
    if ((d<clipmin) || (d>clipmax))
      return 0;
    d=scalar(xp,normal);
    if ((d>cos(viewx)) && (d>cos(viewy)))
      return 0;
  }
  xp=AL*x;
  if (fabs(xp.get(0))>=1e-20) {
    xp=xp/xp.get(0)-target;
    *xl=getmaxx()/2+(int)(SCALEX*scalar(xp,e1));
    *yl=getmaxy()/2-(int)(SCALEY*scalar(xp,e2));
    xp=AR*x;
    if (fabs(xp.get(0))>=1e-20) {
      xp=xp/xp.get(0)-target;
      *xr=getmaxx()/2+(int)(SCALEX*scalar(xp,e1));
      *yr=getmaxy()/2-(int)(SCALEY*scalar(xp,e2));
    } else
      return 0;
  } else
    return 0;
  return 1;
}

//--- monoscopic transformation

int T3D2D(const vector &x,int *x2,int *y2) {
  vector xp;
  double d;

  if (clipping) {
    xp=x-observer;
    d=norm(xp);
    if ((d<clipmin) || (d>clipmax))
      return 0;
    d=scalar(xp,normal);
    if ((d>cos(viewx)) && (d>cos(viewy)))
      return 0;
  }
  xp=A*x;
  if (fabs(xp.get(0))>=1e-20) {
    xp=xp/xp.get(0)-target;
    *x2=getmaxx()/2+(int)(SCALEX*scalar(xp,e1));
    *y2=getmaxy()/2-(int)(SCALEY*scalar(xp,e2));
  } else
    return 0;
  return 1;
}

//--- monoscopic inverse transformation ---------------------

int IT3D2D(int x2,int y2,vector &x) {
  x=target+e1*(double)(x2-getmaxx()/2)/SCALEX+e2*(double)(getmaxy()/2-y2)/SCALEY;
  return 1;
}

//-------------------------------------------------------------------

void SetDrawMode(int dm) {
  dmode=dm;
  if (dmode & DZBUFFERING) {
    if (zbuf!=NULL)
      delete zbuf;
    zbuf=new zbuffer(getmaxx(),getmaxy());
  }
}

int GetDrawMode() { return dmode; }

//--- Drawing section ---------------------------------------

// the drawing takes place on the given default window, the default window is created in the
// xgraph.o module from the libxapp or somewhere else

void putpixel3d(const vector &x,int color) {
  int xl,yl,xr,yr;
  int xm;

  if ((dpymode & DPYMODE)==MONOSCOPICDPY) {
    int x2,y2;

    if (T3D2D(x,&x2,&y2))
      putpixel(x2,y2,color);
  } else if (T3D2D(x,&xl,&yl,&xr,&yr)) {
    int c;
    
    if (dpymode & ANALGLYPHDPY) {
      c=LEFTCOLOR;
      putpixel(xl,yl,c);
      c=RIGHTCOLOR;
      putpixel(xr+1,yr-1,c);
    } else if (dpymode & FRESNELDPY) {
      xm=getmaxx()/2;
      if ((flx+xl<=xm) && (frx+xr>=xm)) {
	putpixel(flx+xl,yl-fly,color);
	putpixel(frx+xr,yr-fry,color);
      }
    } else if (dpymode & EXTERNALDPY) {
      external_putpixel( LEFTEYE,xl,yl,color);
      external_putpixel(RIGHTEYE,xr,yr,color);
    }
  }
}

void line3d(const vector &a,const vector &b,int color) {
  int xl1,yl1,xr1,yr1;
  int xl2,yl2,xr2,yr2;
  int xm;

  if (dpymode==MONOSCOPICDPY) {
    int x1,y1,x2,y2;
    if (T3D2D(a,&x1,&y1) && T3D2D(b,&x2,&y2)) {
      if (color>=0) setcolor(color);
      line(x1,y1,x2,y2);
    }
  } else if (T3D2D(a,&xl1,&yl1,&xr1,&yr1) && T3D2D(b,&xl2,&yl2,&xr2,&yr2)) {
    if ((dpymode & ANALGLYPHDPY)==ANALGLYPHDPY) {
      setcolor(LEFTCOLOR);
      line(xl1,yl1,xl2,yl2);
      setcolor(RIGHTCOLOR);
      line(xr1+1,yr1-1,xr2+1,yr2-1);
    } else if (dpymode & FRESNELDPY) {
      if (color>=0) setcolor(color);
      xm=getmaxx()/2;
      if ((flx+xl1<=xm) && (flx+xl2<=xm) && (frx+xr1>=xm) && (frx+xr2>=xm)) {
	line(flx+xl1,yl1-fly,flx+xl2,yl2-fly);
	line(frx+xr1,yr1-fry,frx+xr2,yr2-fry);
      }
    } else if (dpymode & EXTERNALDPY) {
      external_line( LEFTEYE,xl1,yl1,xl2,yl2,color);
      external_line(RIGHTEYE,xr1,yr1,xr2,yr2,color);
    }
  }
}

void circle3d(const vector m,const vector n,double r,int color) {
}

void rectangle3d(const vector a,const vector b,const vector c,int color) {
  if (dmode & DDRAWING==DWIREFRAME) {
    vector ab,ac,abc;

    ab=a+b;
    ac=a+c;
    abc=a+b+c;
    if (dmode & DDRAWING==DWIREFRAME) {
      line3d(a,ab,color);
      line3d(ab,abc,color);
      line3d(abc,ac,color);
      line3d(ac,a,color);
    }
  } else if (dmode & DDRAWING==DHLINE) {
  } else if (dmode & DDRAWING==DSOLID) {
    if (dmode & DSHADING==DNOSHADING) {
      /* if dmode & DRENDERING==DZBUFFERING draw polygon into z-buffer, with foreground
       border color and background interior */
    } else if (dmode & DSHADING==DFLATSHADE) {
      /* if dmode & DRENDERING==DZBUFFERING draw polygon into z-buffer, with constant color */
    } else if (dmode & DSHADING==DGOURAUDSHADE) {
      /* if dmode & DRENDERING==DZBUFFERING draw polygon into z-buffer, use gouraud-shading */
    } else if (dmode & DSHADING==DRADIOSITY) {
    }
  } else {
  }
}

void cube(const vector a,const vector b,const vector c,const vector d,int color) {
  int l[12];
  int i;
  vector ab,ac,ad,abc,abd,acd,abcd;
  vector n;

  ab=a+b;
  ac=a+c;
  ad=a+d;
  abc=ab+c;
  abd=ab+d;
  acd=ac+d;
  abcd=abc+d;
  
  for (i=0;i<12;i++)
    l[i]=0;

  if ((dmode & DDRAWING==DHLINE) || (dmode & DDRAWING==DSOLID)) {
    n=a+(b+c)/2.0-observer;
    if (scalar(n,d)>=0) {
      if (dmode & DDRAWING==DHLINE) {
	if (!(l[0])) { line3d(a,ab,color); l[0]=1; }
	if (!(l[1])) { line3d(ab,abc,color); l[1]=1; }
	if (!(l[2])) { line3d(abc,ac,color); l[2]=1; }
	if (!(l[3])) { line3d(ac,a,color); l[3]=1; }
      } else
	rectangle3d(a,c,a,color);
    }
    n=ad+(b+c)/2.0-observer;
    if (scalar(n,d)<=0) {
      if (dmode & DDRAWING==DHLINE) {
	if (!(l[8])) { line3d(ad,abd,color); l[8]=1; }
	if (!(l[9])) { line3d(abd,abcd,color); l[9]=1; }
	if (!(l[10])) { line3d(abcd,acd,color); l[10]=1; }
	if (!(l[11])) { line3d(ad,acd,color); l[11]=1; }
      } else
	rectangle3d(ad,b,c,color);
    }
    n=a+(c+d)/2.0-observer;
    if (scalar(n,b)>=0) {
      if (dmode & DDRAWING==DHLINE) {
	if (!(l[4])) { line3d(a,ad,color); l[4]=1; }
	if (!(l[3])) { line3d(ac,a,color); l[3]=1; }
	if (!(l[7])) { line3d(ac,acd,color); l[7]=1; }
	if (!(l[11])) { line3d(ad,acd,color); l[11]=1; }
      } else
	rectangle3d(a,d,c,color);
    }
    n=ab+(c+d)/2.0-observer;
    if (scalar(n,b)<=0) {
      if (dmode & DDRAWING==DHLINE) {
	if (!(l[5])) { line3d(ab,abd,color); l[5]=1; }
	if (!(l[1])) { line3d(ab,abc,color); l[1]=1; }
	if (!(l[6])) { line3d(abc,abcd,color); l[6]=1; }
	if (!(l[9])) { line3d(abd,abcd,color); l[9]=1; }
      } else
	rectangle3d(ab,c,d,color);
    }
    n=a+(b+d)/2.0-observer;
    if (scalar(n,c)>=0) {
      if (dmode & DDRAWING==DHLINE) {
	if (!(l[0])) { line3d(a,ab,color); l[0]=1; }
	if (!(l[5])) { line3d(ab,abd,color); l[5]=1; }
	if (!(l[8])) { line3d(ad,abd,color); l[8]=1; }
	if (!(l[4])) { line3d(a,ad,color); l[4]=1; }
      } else
	rectangle3d(a,b,d,color);
    }
    n=ac+(b+d)/2.0-observer;
    if (scalar(n,c)<=0) {
      if (dmode & DDRAWING==DHLINE) {
	if (!(l[6])) { line3d(abc,abcd,color); l[6]=1; }
	if (!(l[10])) { line3d(abcd,acd,color); l[10]=1; }
	if (!(l[7])) { line3d(ac,acd,color); l[7]=1; }
	if (!(l[2])) { line3d(abc,ac,color); l[2]=1; }
      } else
	rectangle3d(ac,d,b,color);
    }
  } else /*if (dmode & DDRAWING==DWIREFRAME)*/ {
    line3d(a,ab,color);      /*  0 */
    line3d(ab,abc,color);    /*  1 */
    line3d(abc,ac,color);    /*  2 */
    line3d(ac,a,color);      /*  3 */

    line3d(a,ad,color);      /*  4 */
    line3d(ab,abd,color);    /*  5 */
    line3d(abc,abcd,color);  /*  6 */
    line3d(ac,acd,color);    /*  7 */

    line3d(ad,abd,color);    /*  8 */
    line3d(abd,abcd,color);  /*  9 */
    line3d(abcd,acd,color);  /* 10 */
    line3d(acd,ad,color);    /* 11 */
  }
}

void fillrectangle3d(const vector a,const vector b,int color) {
}

void poly3d(int n,vector *pp,int color) {
}

void fillpoly3d(int n,vector *pp,int color) {
}

/*--- z-buffer ----------------------------------------------------*/

zbuffer::zbuffer(unsigned int w,unsigned int h) {
  unsigned int i;

  if ((w<=0) || (h<=0)) {
    printf("zbuffer::zbuffer(%d,%d): dimensio <= 0 not allowed, terminating\n",w,h);
    exit(-1);
  } else if (h>0) {
    width=w;
    height=h;
    buffer=new zpixel* [height];
    for (i=0;i<height;i++)
      buffer[i]=new zpixel [width];
  }
}

zbuffer::~zbuffer() {
  unsigned int i;

  for (i=0;i<height;i++)
    delete buffer[i];
  delete buffer;
}

void DrawPoly(zbuffer &buf,
	      unsigned int n,unsigned int *x,unsigned int *y,unsigned int *z,
	      unsigned int *color) {
  unsigned int minx=MAXINT,maxx=-minx,miny=minx,maxy=maxx,minz=minx,maxz=maxx;
  unsigned int i,j;
  
  for (i=0;i<n;i++) {
    if (minx>x[i]) minx=x[i];
    if (miny>y[i]) miny=y[i];
    if (minz>z[i]) minz=z[i];
    if (maxx<x[i]) maxx=x[i];
    if (maxy<y[i]) maxy=y[i];
    if (maxz<z[i]) maxz=z[i];
  }
  
};

void clear(zbuffer &buf) {
  unsigned int i,j;

  for (i=0;i<buf.height;i++)
    for (j=0;j<buf.width;j++) {
      buf.buffer[i][j].depth=0;
      buf.buffer[i][j].color=0;
    }
}

void flush(zbuffer &buf) {
  unsigned int i,j;

  for (i=0;i<buf.height;i++)
    for (j=0;j<buf.width;j++)
      putpixel(j,i,buf.buffer[i][j].color);
}

zbuffer *Get_zbuffer() { return zbuf; }
void Set_zbuffer(zbuffer *buf) { zbuf=buf; }

void flush3d() {
  if ((dmode & DRENDERING==DZBUFFERING) && (dmode & DDRAWING==DSOLID))
    flush(*zbuf);
//  if (dpymode & EXTERNALDPY)
    external_flush();
}

/*
typedef struct {
  unsigned int n;
  vector *poly;
} zbuffer_struct;

static unsigned long zbuffer_max=4096;
static unsigned long zbuffer_n=0;
static zbuffer_struct *zbuffer=NULL;

void zbufferSetDepth(unsigned long max) {
  zbuffer_struct *p;

  p=zbuffer;
  zbuffer=new zbuffer_struct [max];
  memcpy(zbuffer,p,zbuffer_n*sizeof(zbuffer_struct));
  zbuffer_max=max;
}

void zbufferClear() {
  zbuffer_n=0;
}

void zbufferAdd(unsigned int n,vector *poly) {
  zbuffer[zbuffer_n].n   =n;
  zbuffer[zbuffer_n].poly=poly;
  zbuffer_n++;
}

void zbufferFlush() {
  if (zbuffer_n>0) {
    if ((dmode==DFLATSHADE) || (dmode==DGOURAUDSHADE)) {
    }
  }
}
*/
