
/*--------------------------------------------------------------------
  File:        graphics.cpp
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Compatibility module to Borland's GRAPHICS module
  ----------------------------------------------------------------------
  Notes:       replace #include <graphics.h> by #include "graph.h"
               define USE_GL or USE_X depend on what you want to use,
               either GL or X. If USE_X is defined, set the GRAPHICS_DISPLAY
               to the display variable where you want to put your graphics
               window.
  Date:        22.11.1993
  ----------------------------------------------------------------------
*/ 

#define XTFUNCPROTO
#define ANSI

/*--- For recompiling set the USE_X or USE_GL ----------------*/

/* #define USE_GL */
#define USE_X

/*------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define GRAPH_MODULE
#include "graphics.h"
#undef GRAPH_MODULE

#ifdef USE_GL
#include <gl/gl.h>
#include <gl/device.h>
  
  static long winid;
#endif

#ifdef USE_X
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/X.h>

  static unsigned long ColorOfs=0;
  static Colormap ColorTable;
  static XColor   ColorEntry[256];

  static Visual *VisualInfo;

  static GC XAppGraphicsContext;
  static XEvent XAppGraphicsEvent;

  Window GRAPHICS_WINDOW;
  Display *GRAPHICS_DISPLAY;
#endif

int MAXCOLORS;
static XColor dpycolors[256];
static int* colortab;
static unsigned int sizex,sizey;
static int COLOR,BGCOLOR,FILLCOLOR,FILLTYPE,FILLPATTERN;
static int penx,peny;
static int viewport_x,viewport_y,viewport_w,viewport_h,viewport_c;
static int font_w,font_h;

static char buffer[1310720]; /* 1280*1024*256 */

/*--------------------------------------------------------------------------*/

#ifdef USE_X

void ClearEvent(XEvent *event) {
  if (event!=NULL) 
    XNextEvent(GRAPHICS_DISPLAY,event);
}

void CheckEvent(XGraphicsEventHandler handler) {
  XEvent event;
  int x,y;
  unsigned int border,depth;
  Window rootw;

  if (XCheckWindowEvent(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,
			ButtonPressMask | ButtonReleaseMask | PointerMotionMask | KeyPressMask |
			ColormapChangeMask | ExposureMask | ResizeRedirectMask | LeaveWindowMask |
			EnterWindowMask,
			&event)) {

    if ((event.type!=ButtonPress) && (event.type!=KeyPress)) {
      XGetGeometry(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,&rootw,&x,&y,&sizex,&sizey,&border,&depth);
    }
    handler(&event);
  }
}

int keypressed() {
  int ret;
  XEvent event;

  if (XCheckWindowEvent(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,KeyPressMask,&event)) {
    ret=1;
    XPutBackEvent(GRAPHICS_DISPLAY,&event);
  } else
    ret=0;
  return ret;
}

char readkey() {
  XEvent event;
  char buffer;
  KeySym kbuffer;

  while (!XCheckWindowEvent(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,KeyPressMask,&event));
  XLookupString(&(event.xkey),&buffer,1,&kbuffer,NULL);
  return buffer;
}

int ConvRGBtoPIX(int red,int green,int blue) {
  XColor q;
  int i;
  long dist,min=256*256*3+1;
  int c;

  for (i=0;i<256;i++) {
    q=dpycolors[i];
    dist=abs(q.red-red)+abs(q.green-green)+abs(q.blue-blue);
    if (dist<min) {
      min=dist;
      c=i;
    }
  }
  return(c);
}
#endif

void setrgbpalette(int colornum, int red, int green, int blue) {
  #ifdef USE_GL
    mapcolor(colornum,red,green,blue);
  #endif
  #ifdef USE_X  
    ColorEntry[colornum+ColorOfs].pixel  =(unsigned long)(colornum+ColorOfs);
    ColorEntry[colornum+ColorOfs].red    =(unsigned short)(red*256);
    ColorEntry[colornum+ColorOfs].green  =(unsigned short)(green*256);
    ColorEntry[colornum+ColorOfs].blue   =(unsigned short)(blue*256);
    ColorEntry[colornum+ColorOfs].flags  = DoRed | DoGreen | DoBlue;    
    XStoreColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[colornum+ColorOfs]);
  #endif
}

void SetColors_to_VGA() {
  setrgbpalette(BLACK       ,  0,  0,  0);
  setrgbpalette(BLUE        ,  0,  0,127);
  setrgbpalette(GREEN       ,  0,127,  0);
  setrgbpalette(CYAN        ,  0,127,127);
  setrgbpalette(RED         ,127,  0,  0);
  setrgbpalette(MAGENTA     ,127,  0,127);
  setrgbpalette(BROWN       ,127,127,  0);
  setrgbpalette(LIGHTGRAY   ,127,127,127);
  setrgbpalette(DARKGRAY    , 63, 63, 63);
  setrgbpalette(LIGHTBLUE   ,  0,  0,255);
  setrgbpalette(LIGHTGREEN  ,  0,255,  0);
  setrgbpalette(LIGHTCYAN   ,  0,255,255);
  setrgbpalette(LIGHTRED    ,255,  0,  0);
  setrgbpalette(LIGHTMAGENTA,255,  0,255);
  setrgbpalette(YELLOW      ,255,255,  0);
  setrgbpalette(WHITE       ,255,255,255);
}

void cleardevice() {
  #ifdef USE_GL
    color(BLACK);
    clear();
  #endif
  #ifdef USE_X
    XClearWindow(GRAPHICS_DISPLAY,GRAPHICS_WINDOW);
  #endif
}

#ifdef USE_X
void XGraphicsOn(char *dpy) {
  if (dpy==NULL)
    GRAPHICS_DISPLAY=XOpenDisplay((char*)NULL);
  else
    GRAPHICS_DISPLAY=XOpenDisplay(dpy);
}

void XGraphicsOff() {
  XCloseDisplay(GRAPHICS_DISPLAY);
}
#endif

/*--------------------------------------------------------------------------*/

#ifdef USE_X
  Window initgraph(int *d,int *m,char *path) {
#else
  long int initgraph(int *d,int *m,char *path) {
#endif
  char driver[40];
  #ifdef USE_X
    int i;
    Window root;
    int depth,screen;
    unsigned int border_width = 0;
    int x=0,y=0; 
    XColor q;

    XGCValues GCValues;
    XSetWindowAttributes Attr;
  #endif

  printf("graphics Module V 1.1, (C) Nils Schmeisser, 30.7.1997\n");

  switch (*d) {
   case VGA: switch (*m) {
               case VGAFREE:  sscanf(path,"%d*%d*%d",&sizex,&sizey,&MAXCOLORS);
                              sprintf(driver,"Free VGA %s",path);
			      break;
               case VGAHI:    sizex=640;
                              sizey=480;
                              MAXCOLORS=16;
                              strcpy(driver,"Standard VGA 640*480*16");
                              break;
               case VGAHI256: sizex=640;
                              sizey=480;
                              MAXCOLORS=256;
                              strcpy(driver,"Super VGA 640*480*256");
                              break;
               case   SVGA:   sizex=800;
                              sizey=600;
                              MAXCOLORS=16;
                              strcpy(driver,"Super VGA 800*600*16");
                              break;
               case SVGA256:  sizex=800;
                              sizey=600;
                              MAXCOLORS=256;
                              strcpy(driver,"Super VGA 800*600*256");
                              break;
               case SVGATRUE: sizex=800;
                              sizey=600;
                              MAXCOLORS=16200;   /* Means 16.2 Million, but MAXINT is about 32000 */
                              strcpy(driver,"Super VGA 800*600*16.2 Mill");
                              break;
	     }
             break;

  }
  viewport_x=0;
  viewport_y=0;
  viewport_w=sizex;
  viewport_h=sizey;
  viewport_c=1;

  #ifdef USE_GL
    prefposition(10,long(sizex)+10,1000-long(sizey),1000);
    winid=winopen(driver);
    if ((*m)==SVGATRUE) {
      RGBmode();
      gconfig();
    }
    SetColors_to_VGA();
  #endif
  #ifdef USE_X
    XGraphicsOn(NULL);
    screen=XDefaultScreen(GRAPHICS_DISPLAY);
    root  =XDefaultRootWindow(GRAPHICS_DISPLAY);
    depth =XDefaultDepth(GRAPHICS_DISPLAY,screen);
    VisualInfo=XDefaultVisual(GRAPHICS_DISPLAY,screen);
    /*
    xfonts=(XFontStruct**)malloc(sizeof(XFontStruct*)*256);
    fontlist=XListFontsWithInfo(GRAPHICS_DISPLAY,"*",256,&numfonts,xfonts);
    for (i=0;i<numfonts;i++) {
      printf("%3d: %s\n",i,xfonts[i]);
    }
    */
    font_h=8;
    font_w=6;
    /*
      xfs=XQueryFont(GRAPHICS_DISPLAY,None);
      font_h=xfs->ascent+xfs->descent;
      font_w=XTextWidth(xfs,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
      72)/72;
    */

    // if (MAXCOLORS>216) MAXCOLORS=216;
    ColorTable=XDefaultColormap(GRAPHICS_DISPLAY,screen);

    for (i=0;i<MAXCOLORS;i++) {
      q.pixel=(unsigned long)i;
      XQueryColor(GRAPHICS_DISPLAY,ColorTable,&q);
      ColorEntry[i]=q;
      dpycolors[i].red  =q.red  /256;
      dpycolors[i].green=q.green/256;
      dpycolors[i].blue =q.blue /256;
    }
/*
    ColorTable=XCreateColormap(GRAPHICS_DISPLAY,root,VisualInfo,AllocAll);
    for (i=40;i<256;i++) {
      ColorEntry[i].pixel=(unsigned long)i;
      ColorEntry[i].flags=DoRed | DoGreen | DoBlue;
      ColorEntry[i].red  =0;
      ColorEntry[i].green=0;
      ColorEntry[i].blue =0;
      XAllocColor(GRAPHICS_DISPLAY,ColorTable,&ColorEntry[i]);
    }
  
    XStoreColors(GRAPHICS_DISPLAY,ColorTable,ColorEntry,256);
    SetColors_to_VGA();
 */

    colortab=(int*)malloc(sizeof(int)*MAXCOLORS);

    colortab[BLACK       ]=ConvRGBtoPIX(  0,  0,  0);
    colortab[BLUE        ]=ConvRGBtoPIX(  0,  0,127);
    colortab[GREEN       ]=ConvRGBtoPIX(  0,127,  0);
    colortab[CYAN        ]=ConvRGBtoPIX(  0,127,127);
    colortab[RED         ]=ConvRGBtoPIX(127,  0,  0);
    colortab[MAGENTA     ]=ConvRGBtoPIX(127,  0,127);
    colortab[BROWN       ]=ConvRGBtoPIX(127,127,  0);
    colortab[LIGHTGRAY   ]=ConvRGBtoPIX(127,127,127);
    colortab[DARKGRAY    ]=ConvRGBtoPIX( 63, 63, 63);
    colortab[LIGHTBLUE   ]=ConvRGBtoPIX(  0,  0,255);
    colortab[LIGHTGREEN  ]=ConvRGBtoPIX(  0,255,  0);
    colortab[LIGHTCYAN   ]=ConvRGBtoPIX(  0,255,255);
    colortab[LIGHTRED    ]=ConvRGBtoPIX(255,  0,  0);
    colortab[LIGHTMAGENTA]=ConvRGBtoPIX(255,  0,255);
    colortab[YELLOW      ]=ConvRGBtoPIX(255,255,  0);
    colortab[WHITE       ]=ConvRGBtoPIX(255,255,255);

    Attr.event_mask=ResizeRedirectMask | ExposureMask | ButtonPressMask | ColormapChangeMask | 
                    LeaveWindowMask | EnterWindowMask;
    Attr.background_pixel=0;
    Attr.border_pixel    =1;
    Attr.colormap  =ColorTable;
    GRAPHICS_WINDOW=XCreateWindow(GRAPHICS_DISPLAY,root,
                                  x,y,sizex,sizey,border_width,depth,
                                  InputOutput,VisualInfo,
                                  CWBackPixel | CWEventMask | CWBorderPixel | CWColormap,&Attr);
  
    GCValues.function  =GXcopy;
    GCValues.plane_mask=AllPlanes;
    GCValues.foreground=1;
    GCValues.background=0;
    XAppGraphicsContext=XCreateGC(GRAPHICS_DISPLAY, root /* GRAPHICS_WINDOW */ ,
                                  GCFunction | GCPlaneMask | GCForeground | GCBackground,&GCValues); 
    XMapWindow(GRAPHICS_DISPLAY,GRAPHICS_WINDOW);
    /*    XInstallColormap(GRAPHICS_DISPLAY,ColorTable);*/
    XSelectInput(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,
                 ButtonPressMask | ButtonReleaseMask | PointerMotionMask | KeyPressMask |
                 ColormapChangeMask | ExposureMask | LeaveWindowMask | EnterWindowMask);
    XNextEvent(GRAPHICS_DISPLAY,&XAppGraphicsEvent);
    /*
      for (i=0;i<256;i++) {
      q.pixel=(unsigned long)i;
      XQueryColor(GRAPHICS_DISPLAY,ColorTable,&(dpycolors[i]));
      dpycolors[i].red   /= 256;
      dpycolors[i].green /= 256;
      dpycolors[i].blue  /= 256;
      }
    */
  #endif
  COLOR=WHITE;
  BGCOLOR=BLACK;  
  cleardevice();
  #ifdef USE_X
    return GRAPHICS_WINDOW;
  #else
    return winid;
  #endif
}

void moveto(int x,int y) {
  penx=x;
  peny=y;
}

void lineto(int x,int y) {
  line(penx,peny,x,y);
  penx=x;
  peny=y;
}

/*-----------------------------------------------------------------------*/

void arc(int x,int y,int r,int startangle,int stopangle) {
  XDrawArc(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
	   x-r+viewport_x,y-r+viewport_y,2*r,2*r,startangle*64,stopangle*64);
}

void pie(int x,int y,int r,int startangle,int stopangle) {
  XFillArc(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
	   x-r+viewport_x,y-r+viewport_y,2*r,2*r,startangle*64,stopangle*64);
}

/*-----------------------------------------------------------------------*/

void circle(int x,int y,int r) { arc(x,y,r,0,360); }

void closegraph() {
  #ifdef USE_GL
    winclose(winid);
  #endif
  #ifdef USE_X
    free(colortab);
    XUninstallColormap(GRAPHICS_DISPLAY,ColorTable);
    XFreeColormap(GRAPHICS_DISPLAY,ColorTable);
    XFreeGC(GRAPHICS_DISPLAY,XAppGraphicsContext);
    XDestroyWindow(GRAPHICS_DISPLAY,GRAPHICS_WINDOW);	
    XGraphicsOff();
  #endif
}

int getmaxcolor() {
  return(MAXCOLORS);
}

int getmaxx() {
  return sizex-1;
}

int getmaxy() {
  return sizey-1;
}

int getvmaxx() {
  return viewport_w-1;
}

int getvmaxy() {
  return viewport_h-1;
}

void setviewport(int x1,int y1,int x2,int y2,int clip) {
  viewport_x=x1;
  viewport_y=y1;
  viewport_w=x2-x1+1;
  viewport_h=y2-y1+1;
  viewport_c=clip;
}

void clearviewport() {
  int fc,fg;

  fg=COLOR;
  fc=FILLCOLOR;

  COLOR=BGCOLOR;
  FILLCOLOR=BGCOLOR;
  bar(0,0,getvmaxx(),getvmaxy());

  setcolor(fg);
  COLOR=fg;
  FILLCOLOR=fc;
}

void setcolor(int color) {
  COLOR=colortab[color];
  XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,COLOR);
}

void setbgcolor(int color) {
  BGCOLOR=colortab[color];
}

void putpixel(int x,int y,int col) {
  #ifdef USE_GL
    color(col);
    rect(x+viewport_x,getmaxy()-y-viewport_y,x+1+viewport_x,getmaxy()-y-1-viewport_y);
  #endif
  #ifdef USE_X
    if ((!viewport_c) || 
	(viewport_c && (x>=0) && (x<viewport_w) && (y>=0) && (y<viewport_h))) {

    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,col+ColorOfs);
    XDrawPoint(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
	       x+viewport_x,y+viewport_y);
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,COLOR);

    }
  #endif
}

void line(int x1,int y1,int x2,int y2) {
  long p1[2],p2[2];

  #ifdef USE_GL
    color(COLOR);
    p1[0]=x1+viewport_x;
    p1[1]=getmaxy()-y1-viewport_y;
    p2[0]=x2+viewport_x;
    p2[1]=getmaxy()-y2-viewport_y;
    bgnline();
      v2i(p1);
      v2i(p2);
    endline();
  #endif
  #ifdef USE_X
    XDrawLine(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
	      x1+viewport_x,y1+viewport_y,x2+viewport_x,y2+viewport_y);	
  #endif
}

void putRGBpixel(int x,int y,int r,int g,int b) {
  #ifdef USE_GL
    RGBcolor(r,g,b);
    rect(x+viewport+x,getmaxy()-y-viewport_y,x+1+viewport_x,getmaxy()-y-1-viewport_y);
  #endif
  #ifdef USE_X
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,ConvRGBtoPIX(r,g,b));
    XDrawPoint(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,x+viewport_x,y+viewport_y);
    XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,COLOR);
  #endif
}

unsigned long imagesize(int x1,int y1,int x2,int y2) {
  unsigned long s;

  s=(unsigned long)(x2-x1+1)*(unsigned long)(y2-y1+1)+4;
  return s;
}

int textheight() {
  /*  XFontStruct* xfs;

  xfs=XQueryFont(GRAPHICS_DISPLAY,None);
  font_h=xfs->ascent+xfs->descent;
  */
  return font_h;
}

int textwidth() {
  /*  XFontStruct* xfs;

  xfs=XQueryFont(GRAPHICS_DISPLAY,None);
  font_w=XTextWidth(xfs,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",72)/72;
  */
  return font_w;
}

void outtextxy(int x,int y,char *s) {
  XTextItem text;

  text.chars=s;
  text.nchars=strlen(s);
  text.delta=1;
  text.font=None;
  XDrawText(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
	    x+viewport_x,y+viewport_y+font_h,&text,1);
}

void setfillstyle(int ftype,int fpattern,int fcolor) {
  FILLTYPE   =ftype;
  FILLPATTERN=fpattern;
  FILLCOLOR  =fcolor;
}

void bar(int x0,int y0,int x1,int y1) {
  XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,FILLCOLOR);
  XFillRectangle(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
		 x0+viewport_x,y0+viewport_y,x1-x0,y1-y0);
  XSetForeground(GRAPHICS_DISPLAY,XAppGraphicsContext,COLOR);
  XDrawRectangle(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
		 x0+viewport_x,y0+viewport_y,x1-x0,y1-y0);
}

void rectangle(int x0,int y0,int x1,int y1) {
  XDrawRectangle(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,
		 x0+viewport_x,y0+viewport_y,x1-x0,y1-y0);
}

void getimage(int x1,int y1,int x2,int y2,void *bitmap) {
  unsigned short w,h,*d;
  char *data;
  #ifdef USE_X
    XImage* p;
    unsigned long i,j,o;
  #endif

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data++;
  data++;
  data++;
  data++;

  w=x2-x1+1;
  h=y2-y1+1;
  (*d)=w;
  d++;
  (*d)=h;

  #ifdef USE_GL
    Int16 *dat;
    long i,j;

    dat=new Int16 [w*h];
    rectread(x1,y1,x2,y2,dat);
    for (j=h-1;j>=0;j--) {
      for (i=0;i<w;i++) {
        data[((h-1)-j)*w+i]=(char)dat[j*w+i];
      }
    }
    delete dat;
  #endif
  #ifdef USE_X

    p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
    (*p)=*XGetImage(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,x1,y1,w,h,0xffffffff,ZPixmap);
    o=p->bytes_per_line-w;
    for (i=0;i<h;i++) for (j=0;j<w;j++) data[i*w+j]=p->data[i*(w+o)+j];
  #endif
}

void putimage(int x,int y,void *bitmap,int flag) {
  unsigned short w,h,*d;
  char *data;
  #ifdef USE_X
    XImage* p;
  #endif

  data=(char*)bitmap;
  d=(unsigned short*)bitmap;
  data++;
  data++;
  data++;
  data++;
  w=(*d);
  d++;
  h=(*d);

  #ifdef USE_GL
    Int16 *dat;
    long i,j;

    dat=new Int16 [w*h];
    for (j=h-1;j>=0;j--) {
      for (i=0;i<w;i++) {
        dat[j*w+i]=(Int16)data[((h-1)-j)*w+i];
      }
    }
    rectwrite(x,y,x+w-1,y+h-1,dat);
    delete dat;
  #endif
  #ifdef USE_X
    p=XCreateImage(GRAPHICS_DISPLAY,VisualInfo,8,ZPixmap,0,data,w,h,8,w);
    XPutImage(GRAPHICS_DISPLAY,GRAPHICS_WINDOW,XAppGraphicsContext,p,
	      0,0,x+viewport_x,y+viewport_y,w,h);
  #endif
}

void flush() {
  XFlush(GRAPHICS_DISPLAY);
}
