/*--------------------------------------------------------------------
  File:        graphics.h
  Author:      Nils Schmeisser
  ----------------------------------------------------------------------
  Description: Compatibility module to Borland's GRAPHICS module, include
  ----------------------------------------------------------------------
  Notes:       replace #include <graphics.h> by #include "graph.h"
               define USE_GL or USE_X depend on what you want to use,
               either GL or X, in "deflib.h".
               If USE_X is defined, set the GRAPHICS_DISPLAY to the 
               display variable where you want to put your graphics
               window.
  Date:        22.11.1993
  ----------------------------------------------------------------------
*/ 

#ifndef GRAPHICS_
#define GRAPHICS_

#define XTFUNCPROTO
#define ANSI

#include <stdlib.h>

/* Graphics drivers */

#define DETECT   0
#define CGA      1
#define MCGA     2
#define EGA      3
#define EGA64    4
#define EGAMONO  5
#define IBM8514  6
#define HERCMONO 7
#define ATT400   8
#define VGA      9
#define PC3270  10

/* Graphics modes */

#define CGAC0    0/* 320*200 palette 0 */
#define CGAC1    1/* 320*200 palette 1 */
#define CGAC2    2/* 320*200 palette 2 */
#define CGAC3    3/* 320*200 palette 3 */
#define CGAHI    4/* 640*200*2 */

#define MCGAC0   0/* 320*200 palette 0 */
#define MCGAC1   1/* 320*200 palette 1 */
#define MCGAC2   2/* 320*200 palette 2 */
#define MCGAC3   3/* 320*200 palette 3 */
#define MCGAMED  4/* 640*200*2 */
#define MCGAHI   5/* 640*480*2 */

#define VGAFREE  1 /* free definable */
#define VGAHI    2 /* 640*480*16 */
#define VGAHI256 3 /* 640*480*256 */
#define SVGA     4 /* 800*600*16 */
#define SVGA256  5 /* 800*600*256 */
#define SVGATRUE 6 /* 800*600*16.2Mill */

#define False    0
#define True     1
#define ClipOn   1

#define BLACK         0
#define BLUE          1
#define GREEN         2
#define CYAN          3
#define RED           4
#define MAGENTA       5
#define BROWN         6
#define LIGHTGRAY     7
#define DARKGRAY      8
#define LIGHTBLUE     9
#define LIGHTGREEN   10
#define LIGHTCYAN    11
#define LIGHTRED     12
#define LIGHTMAGENTA 13
#define YELLOW       14
#define WHITE        15

#ifdef USE_X
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/X.h>

extern Display *GRAPHICS_DISPLAY;
extern Window GRAPHICS_WINDOW;

typedef void (*XGraphicsEventHandler)(XEvent *event);
extern void ClearEvent(XEvent *event);
extern void CheckEvent(XGraphicsEventHandler handler);
#endif

extern int MAXCOLORS;

extern void XGraphicsOn(char *dpy);
extern void XGraphicsOff();
#ifdef USE_X
extern Window initgraph(int *d,int *m,char *path);
#else
extern long int initgraph(int *d,int *m,char *path);
#endif

extern void moveto(int x,int y);
extern void lineto(int x,int y);
extern void arc(int x,int y,int r,int startangle,int stopangle);
extern void pie(int x,int y,int r,int startangle,int stopangle);
extern void circle(int x,int y,int r);
extern void closegraph();
extern void cleardevice();

extern void setviewport(int x1,int y1,int x2,int y2,int clip);
extern void clearviewport();
extern int getmaxx();
extern int getmaxy();
extern int getvmaxx();
extern int getvmaxy();

extern void setrgbpalette(int colornum, int red, int green, int blue);
extern void setcolor(int color);
extern void setbgcolor(int color);
extern int getmaxcolor();

extern void putpixel(int x,int y,int col);
extern void line(int x1,int y1,int x2,int y2);
extern void putRGBpixel(int x,int y,int r,int g,int b);

extern unsigned long imagesize(int x1,int y1,int x2,int y2);
extern void getimage(int x1,int y1,int x2,int y2,void *bitmap);
extern void putimage(int x,int y,void *bitmap,int flag);

extern int textwidth();
extern int textheight();
extern void outtextxy(int x,int y,char *s);
extern void setfillstyle(int ftype,int fpattern,int fcolor);
extern void bar(int x0,int y0,int x1,int y1);
extern void rectangle(int x0,int y0,int x1,int y1);

extern int keypressed();
extern char readkey();
extern void flush();

#endif
