/* 
   linux/drivers/char/lcdshutter.c
   Copyright (C) 1996 Nils Schmeisser
   Generic LCDShutter driver for Linux running on an IBM compatible computer.
*/

#include "lcdshutter.h"
#include "release.h"

#ifdef __cplusplus
  extern "C" {
#endif
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/module.h>
#ifdef __cplusplus
  }
#endif

static int LCDSHUTTER_PORT;

/* get_timer0():
   returns the current value of timer 0. This is a 16 bit counter that starts
   at LATCH and counts down to 0 */

inline int get_timer0 (void) {
  int t0, t1;
  outb (0, PIT_MODE);
  t0 = (int) inb (PIT_COUNTER_0);
  t1 = ((int) inb (PIT_COUNTER_0) << 8) + t0;
  return (t1);
}

inline int ControlLCDShutter(int cmd,int device) {
  if (   (cmd==LCDSHUTTER_CLEAR) || (cmd==LCDSHUTTER_CLOSE)
      || (cmd==LCDSHUTTER_OLEFT) || (cmd==LCDSHUTTER_ORIGHT)) {
    switch (device) {
      /*--- the CyberBoy --------------------------*/
      case 0: outb(cmd,LCDSHUTTER_PORT); break;

      /*--- the Sega glasses ----------------------*/
      case 1: break;
    }	
    /* should delay now for a certain time to make sure, the glasses
       have reacted */

    return 0;
  } else
    return -EINVAL;
}

static int lcdshutter_ioctl (struct inode *inode,
			     struct file *file,
			     unsigned int cmd,
			     unsigned long arg) {
  unsigned int minor;

  #ifdef DEBUG
    printk("LCDShutter ioctl\n");
  #endif

  if (MAJOR (inode->i_rdev) != LCDSHUTTER_MAJOR)
    return -EINVAL;
  minor=MINOR(inode->i_rdev);
  if (minor >= LCDSHUTTER_MAX)
    return -ENODEV;
  return ControlLCDShutter(cmd,minor);
}

static int lcdshutter_open(struct inode *inode, struct file *file) {
  unsigned int minor=MINOR(inode->i_rdev);

  if (minor>=LCDSHUTTER_MAX)
    return -ENODEV;	 /*check for /dev/lcd* */
  MOD_INC_USE_COUNT;
  switch (minor) {
    case 0: LCDSHUTTER_PORT=CYBERBOY_PORT; break;
    case 1: LCDSHUTTER_PORT=SEGA_PORT; break;
  }
  return 0;
}


static void lcdshutter_release (struct inode *inode, struct file *file) {

  inode->i_atime=CURRENT_TIME;
  MOD_DEC_USE_COUNT;
}

static int lcdshutter_write (struct inode *inode,
			     struct file *file, char *buf, int count) {
  return 0;
}

static int lcdshutter_read (struct inode *inode,
			    struct file *file, char *buf, int count) {
  return 0;
}

static struct file_operations lcdshutter_fops =
{
	NULL,			/* js_lseek*/
	lcdshutter_read,	/* js_read */
	lcdshutter_write,	/* js_write*/
	NULL,			/* js_readaddr*/
	NULL,			/* js_select */
	lcdshutter_ioctl,	/* js_ioctl*/
	NULL,			/* js_mmap */
	lcdshutter_open,	/* js_open*/
	lcdshutter_release,     /* js_release*/
	NULL			/* js_sync */
};

#ifdef __cplusplus
extern "C" {
#endif

int init_module(void) {

  if (register_chrdev(LCDSHUTTER_MAJOR,"lcdshutter",&lcdshutter_fops)) {
    printk ("Unable to get major=%d for lcdshutter\n",LCDSHUTTER_MAJOR);
    return 1;
  } else
    return 0;
}

void cleanup_module (void) {
  if (MOD_IN_USE)
    printk ("LCDShutter: device busy, remove delayed.\n");
  if (unregister_chrdev (LCDSHUTTER_MAJOR, "LCDShutter") != 0) {
    printk ("cleanup_module failed.\n");
  } else {
    printk ("cleanup_module succeeded.\n");
  }
}

#ifdef __cplusplus
}
#endif




