#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

extern "C" {
  #include <graphics/graphics.h>
  #include <lcdapi-1.0/lcdshutter.h>
}

static unsigned long DELAYCYC;    /*--- cycles per msec ---*/

void delay(unsigned long msec) {
  unsigned long i,n;

  n=DELAYCYC*msec;
  for (i=0;i<n;i++);
}

int readln(FILE *f,char* s) {
  char c;
  int i,result;

  i=0;
  do {
    result=fscanf(f,"%c",&c);
    if (result!=EOF) {
      if (c!='\n') {
	s[i]=c;
	i++;
      }
    }
  } while ((result!=EOF) && (c!='\n'));
  s[i]=0;
  return result;
}

double distance;

void rotate2d(double &x,double &y,double u) {
  double xn,yn,su,cu;
  
  su=sin(u);
  cu=cos(u);

  xn=x*cu-y*su;
  yn=x*su+y*cu;
  x=xn;
  y=yn;
}

void rotate(double &x,double &y,double &z,double u,double v,double w) {
  rotate2d(y,z,u);
  rotate2d(z,x,v);
  rotate2d(x,y,w);
}
/*
int X2D(double x,double y,double z,double scale) {

  return (int)(scale*(x*distance/(y-distance)))+getmaxx()/2;
}

int Y2D(double x,double y,double z,double scale) {
  return getmaxy()/2-(int)(scale*(z*distance/(y-distance)));
}
*/

int X2D(double x,double y,double z,double xb,double yb,double scale) {
  return (int)(scale*((y*xb-x*yb)/(y-yb)))+getmaxx()/2;
}

int Y2D(double x,double y,double z,double xb,double yb,double scale) {
  return getmaxy()/2-(int)(scale*(z*yb/(yb-y)));
}

int natom,ratom;
double *x,*y,*z;
int nvalenz;
int **valenz;
int *cval;
double xb[2],yb[2];
int b;

int main(int argc,char **argv) {
  FILE *f;
  int i;
  int d,m;
  int x1,y1,x2,y2,i1,i2;
  double scale;
  double u,v,w,xn,yn,zn,dw;
  char c;
  char s[256];
  unsigned long isize;
  void* image[2];
  unsigned long n;
  clock_t c1,c2,dc;
  int tlcd,tcrt,ti,lcd;
  int numbering;

  printf("Molecule Viewer V 1.0, (C) Nils Schmeisser, 1997\n\n");
  printf("View controls:  1 2: rotate around x-axis\n");
  printf("                4 6: rotate around y-axis\n");
  printf("                7 9: rotate around z-axis\n");
  printf("                - +: zoom out, zoom in\n");
  printf("                  n: toggle atom numbering on/off\n");
  printf("                ESC: exit\n");
  printf("                any other key updates the screen\n");

  numbering=0;
  lcd=0;
  tlcd=2;
  tcrt=3;
  ti  =7;
  dw=5*M_PI/180;
  f=NULL;
  for (i=1;i<argc;i++) {
    if (!(strcmp(argv[i],"-f")))
      f=fopen(argv[i+1],"r");
    if (!(strcmp(argv[i],"+lcd")))
      lcd=1;
    if (!(strcmp(argv[i],"+num")))
      numbering=1;
    if (!(strcmp(argv[i],"-tlcd")))
      sscanf(argv[i+1],"%d",&tlcd);
    if (!(strcmp(argv[i],"-tcrt")))
      sscanf(argv[i+1],"%d",&tcrt);
    if (!(strcmp(argv[i],"-ti")))
      sscanf(argv[i+1],"%d",&ti);
    if (!(strcmp(argv[i],"-help"))) {
      printf("Usage: mviewer [[+lcd] [-tlcd lcd_t] [-tcrt crt_t] [-ti img_t]] [+num] -f file\n");
      printf("       +lcd     enables CyberBoy Shutter glasses (default: off)\n");
      printf("       -tlcd    reaction time for shutter glasses (default: 2 ms)\n");
      printf("       -tcrt    reaction time of crt display (default: 3 ms)\n");
      printf("       -ti      time to keep subimage (default: 7 ms)\n");
      printf("       +num     set numering of atoms on (default: off)\n");
      printf("       -f file  the data file\n");
      exit(0);
    }
  }
  if (f==NULL) {
    printf("No model file given or file not found, check '-f <file>' option.\n");
    printf("Type 'mviewer -help' for help.\n");
    exit(-1);
  }

  fscanf(f,"%d %d",&natom,&ratom);
  x=new double [natom];
  y=new double [natom];
  z=new double [natom];
  u=v=w=0;

  for (i=0;i<natom;i++) {
    fscanf(f,"%lf %lf %lf",&(x[i]),&(y[i]),&(z[i]));
    x[i]-=2;
    y[i]-=2;
    z[i]-=0.5;
  }

  fscanf(f,"%d",&nvalenz);
  cval=new int [nvalenz];
  valenz=new int* [nvalenz];
  
  for (i=0;i<nvalenz;i++) {
    valenz[i]=new int [2];
    fscanf(f,"%d %d",&(valenz[i][0]),&(valenz[i][1]));
    readln(f,s);
    if (strlen(s)>0) {
      sscanf(s,"%d",&(cval[i]));
    } else
      cval[i]=GREEN;
  }

  fclose(f);

  distance=100;
  yb[0]=yb[1]=-distance;
  if (lcd) {
    xb[0]=-10;
    xb[1]= 10;
  } else
    xb[0]=xb[1]=0;
  scale=1;

  /*--- calibrating delay loop ---*/
  printf("Calibrating delay loop\n");
  DELAYCYC=1;
  dc=0;
  while (dc<10) {
    DELAYCYC*=10;
    c1=clock();
    delay(1);
    c2=clock();
    dc=c2-c1;
  }
  n=DELAYCYC;
  DELAYCYC=(n*CLOCKS_PER_SEC)/dc/1000;
  printf("delay loop: DELAYCYC=%d\n",(unsigned int)DELAYCYC);

  d=VGA;
  m=VGAFREE;
  initgraph(&d,&m,"800*600*256");

  if (lcd) {
    if (ShutterInit()>0) {
      ShutterClear();
      delay(tlcd);
    } else {
      printf("Sorry, LCD shutter device not available, switching to normal mode.\n");
      lcd=0;
    }
  }

  b=0;
  do {
    cleardevice();

    if (lcd) {
      flush();
      delay(tcrt);
      if (b==0)
	ShutterLeft();
      else
	ShutterRight();
      delay(tlcd);
    }

    for (i=0;i<nvalenz;i++) {
      i1=valenz[i][0];
      i2=valenz[i][1];
      xn=x[i1];
      yn=y[i1];
      zn=z[i1];
      rotate(xn,yn,zn,u,v,w);
      x1=X2D(xn,yn,zn,xb[b],yb[b],scale);
      y1=Y2D(xn,yn,zn,xb[b],yb[b],scale);
      xn=x[i2];
      yn=y[i2];
      zn=z[i2];
      rotate(xn,yn,zn,u,v,w);
      x2=X2D(xn,yn,zn,xb[b],yb[b],scale);
      y2=Y2D(xn,yn,zn,xb[b],yb[b],scale);
      
      setcolor(cval[i]);
      line(x1,y1,x2,y2);
    }
    setcolor(WHITE);
    for (i=0;i<natom;i++) {
      xn=x[i];
      yn=y[i];
      zn=z[i];
      rotate(xn,yn,zn,u,v,w);
      x1=X2D(xn,yn,zn,xb[b],yb[b],scale);
      y1=Y2D(xn,yn,zn,xb[b],yb[b],scale);
      pie(x1,y1,ratom,0,360);
      if (numbering) {
	sprintf(s,"%2d",i);
	outtextxy(x1+5,y1+5,s);
      }
    }
    
    sprintf(s,"u=%4d v=%4d w=%4d scale=%4.1lf",
	    (int)(180/M_PI*u),(int)(180/M_PI*v),(int)(180/M_PI*w),scale);
    outtextxy(5,getmaxy()-13,s);
    
    flush();

    if (lcd) {
      delay(ti);
      b=(b + 1) % 2;
    }

    c=0;
    if (lcd) {
      if (keypressed())
	c=readkey();
    } else
      c=readkey();

    if (c!=0) {
      switch (c) {
      case '9': w=w+dw; break;
      case '7': w=w-dw; break;
      case '6': v=v+dw; break;
      case '4': v=v-dw; break;
      case '3': u=u+dw; break;
      case '1': u=u-dw; break;
      case '-': scale=scale/1.5; break;
      case '+': scale=scale*1.5; break;
      case 'n': numbering= !numbering; break;
      }
      if (scale<0.01) scale=0.01;
      if (scale>5000) scale=5000;
      if (u<-M_PI) u=2*M_PI-u;
      if (u> M_PI) u=u-2*M_PI;
      if (v<-M_PI) v=2*M_PI-v;
      if (v> M_PI) v=v-2*M_PI;
      if (w<-M_PI) w=2*M_PI-w;
      if (w> M_PI) w=w-2*M_PI;
    }
  } while (c!=27);
    
  if (lcd) {
    ShutterClear();
    ShutterDone();
  }

  closegraph();

  return 0;
}
